import sys

from nose.plugins.skip import SkipTest

from bugsnag import Configuration
from bugsnag.delivery import (UrllibDelivery, RequestsDelivery,
                              create_default_delivery)

from tests.utils import IntegrationTest


class DeliveryTest(IntegrationTest):

    def setUp(self):
        super(DeliveryTest, self).setUp()
        self.config = Configuration()
        self.config.configure(api_key='abc',
                              endpoint=self.server.address,
                              use_ssl=False,
                              asynchronous=False)

    def test_urllib_delivery_full_url(self):
        self.config.configure(endpoint=self.server.url, use_ssl=None)
        UrllibDelivery().deliver(self.config, '{"legit": 4}')

        self.assertSentReportCount(1)
        request = self.server.received[0]
        self.assertEqual(request['json_body'], {"legit": 4})

    def test_urllib_delivery(self):
        UrllibDelivery().deliver(self.config, '{"legit": 4}')

        self.assertSentReportCount(1)
        request = self.server.received[0]
        self.assertEqual(request['json_body'], {"legit": 4})
        self.assertEqual(request['headers']['Content-Type'],
                         'application/json')

    def test_requests_delivery(self):
        if sys.version_info < (2, 7):
            raise SkipTest("Requests is incompatible with 2.6")

        try:
            import requests  # noqa
        except ImportError:
            raise SkipTest("Requests is not installed")

        RequestsDelivery().deliver(self.config, '{"legit": 4}')

        self.assertSentReportCount(1)
        request = self.server.received[0]
        self.assertEqual(request['json_body'], {"legit": 4})
        self.assertEqual(request['headers']['Content-Type'],
                         'application/json')

    def test_requests_delivery_full_url(self):
        if sys.version_info < (2, 7):
            raise SkipTest("Requests is incompatible with 2.6")

        try:
            import requests  # noqa
        except ImportError:
            raise SkipTest("Requests is not installed")

        self.config.configure(endpoint=self.server.url)
        del self.config.use_ssl
        RequestsDelivery().deliver(self.config, '{"legit": 4}')

        self.assertSentReportCount(1)
        request = self.server.received[0]
        self.assertEqual(request['json_body'], {'legit': 4})

    def test_create_default_delivery(self):
        delivery = create_default_delivery()

        if 'requests' in sys.modules:
            self.assertTrue(isinstance(delivery, RequestsDelivery))
        else:
            self.assertTrue(isinstance(delivery, UrllibDelivery))



import json
import time
import unittest
from threading import Thread

from six.moves import BaseHTTPServer

import bugsnag


class IntegrationTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.server = FakeBugsnagServer()

    def setUp(self):
        self.server.received = []

    def tearDown(self):
        client = bugsnag.Client()
        client.configuration.api_key = 'some key'
        bugsnag.legacy.default_client = client
        bugsnag.legacy.configuration = client.configuration

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()

    def assertSentReportCount(self, count):
        self.assertEqual(len(self.server.received), count)

    def assertExceptionName(self, received_index, event_index, name):
        json_body = self.server.received[received_index]['json_body']
        event = json_body['events'][event_index]
        exception = event['exceptions'][0]
        self.assertEqual(exception['errorClass'], name)


class FakeBugsnagServer(object):
    """
    A server which accepts a single request, recording the JSON payload and
    other request information
    """

    def __init__(self):
        self.received = []
        self.paused = False

        class Handler(BaseHTTPServer.BaseHTTPRequestHandler):

            def do_POST(handler):
                start = time.time()

                while self.paused:
                    if time.time() > (start + 0.5):
                        raise Exception('Paused HTTP server timeout')

                    time.sleep(0.001)

                length = int(handler.headers['Content-Length'])
                raw_body = handler.rfile.read(length).decode('utf-8')
                if handler.path != '/ignore':
                    self.received.append({'headers': handler.headers,
                                          'json_body': json.loads(raw_body),
                                          'path': handler.path,
                                          'method': handler.command})
                handler.send_response(200)
                handler.end_headers()
                return ()

            def log_request(self, *args):
                pass

        self.server = BaseHTTPServer.HTTPServer(('localhost', 0), Handler)
        self.server.timeout = 0.5
        self.thread = Thread(target=self.server.serve_forever, args=(0.1,))
        self.thread.daemon = True
        self.thread.start()

    @property
    def address(self):
        return '{0}:{1}'.format(*self.server.server_address)

    @property
    def url(self):
        return 'http://%s' % self.address

    def shutdown(self):
        self.server.shutdown()
        self.thread.join()
        self.server.server_close()


class ScaryException(Exception):
    pass



from __future__ import division, print_function, absolute_import

import inspect
import six
from json import JSONEncoder
from threading import local as threadlocal

import bugsnag

MAX_PAYLOAD_LENGTH = 128 * 1024
MAX_STRING_LENGTH = 1024


class SanitizingJSONEncoder(JSONEncoder):
    """
    A JSON encoder which handles filtering and conversion from JSON-
    incompatible types to strings.

    >>> encoder = SanitizingJSONEncoder(filters=['bananas'])
    >>> encoder.encode({'carrots': 4, 'bananas': 5})
    "{'carrots': 4, 'bananas': '[FILTERED]'}"
    """

    filtered_value = '[FILTERED]'
    recursive_value = '[RECURSIVE]'
    unencodeable_value = '[BADENCODING]'

    def __init__(self, keyword_filters=None, **kwargs):
        self.filters = list(map(str.lower, keyword_filters or []))
        super(SanitizingJSONEncoder, self).__init__(**kwargs)

    def encode(self, obj):
        safe_obj = self._sanitize(obj, False)
        payload = super(SanitizingJSONEncoder, self).encode(safe_obj)
        if len(payload) > MAX_PAYLOAD_LENGTH:
            safe_obj = self._sanitize(safe_obj, True)
            return super(SanitizingJSONEncoder, self).encode(safe_obj)
        else:
            return payload

    def filter_string_values(self, obj, ignored=None):
        """
        Remove any value from the dictionary which match the key filters
        """
        if not ignored:
            ignored = []

        if id(obj) in ignored:
            return self.recursive_value

        if isinstance(obj, dict):
            ignored.append(id(obj))

            clean_dict = {}
            for key, value in six.iteritems(obj):
                is_string = isinstance(key, six.string_types)
                if is_string and any(f in key.lower() for f in self.filters):
                    clean_dict[key] = self.filtered_value
                else:
                    clean_dict[key] = self.filter_string_values(value, ignored)

            return clean_dict

        return obj

    def default(self, obj):
        """
        Coerce values to strings if possible, otherwise replace with
        '[BADENCODING]'
        """
        try:
            if six.PY3 and isinstance(obj, bytes):
                return six.text_type(obj, encoding='utf-8', errors='replace')
            else:
                return six.text_type(obj)

        except Exception:
            bugsnag.logger.exception('Could not add object to payload')
            return self.unencodeable_value

    def _sanitize(self, obj, trim_strings, ignored=None):
        """
        Replace recursive values and trim strings longer than
        MAX_STRING_LENGTH
        """
        if not ignored:
            ignored = []

        if id(obj) in ignored:
            return self.recursive_value
        elif isinstance(obj, dict):
            ignored.append(id(obj))
            return self._sanitize_dict(obj, trim_strings, ignored)
        elif isinstance(obj, (set, tuple, list)):
            ignored.append(id(obj))
            items = []
            for value in obj:
                items.append(self._sanitize(value, trim_strings, ignored))
            return items
        elif trim_strings and isinstance(obj, six.string_types):
            return obj[:MAX_STRING_LENGTH]
        else:
            return obj

    def _sanitize_dict(self, obj, trim_strings, ignored):
        """
        Trim individual values in an object, applying filtering if the object
        is a FilterDict
        """
        if isinstance(obj, FilterDict):
            obj = self.filter_string_values(obj)

        clean_dict = {}
        for key, value in six.iteritems(obj):
            clean_dict[key] = self._sanitize(value, trim_strings, ignored)

        return clean_dict


class FilterDict(dict):
    """
    Object which will be filtered when encoded
    """
    pass


def fully_qualified_class_name(obj):
    module = inspect.getmodule(obj)
    if module is not None and module.__name__ != "__main__":
        return module.__name__ + "." + obj.__class__.__name__
    else:
        return obj.__class__.__name__


def package_version(package_name):
    try:
        import pkg_resources
    except ImportError:
        return None
    else:
        try:
            return pkg_resources.get_distribution(package_name).version
        except pkg_resources.DistributionNotFound:
            return None


def merge_dicts(lhs, rhs):
    for key, value in rhs.items():
        if isinstance(value, dict):
            node = lhs.setdefault(key, {})
            merge_dicts(node, value)
        elif isinstance(value, list):
            array = lhs.setdefault(key, [])
            array += value
        else:
            lhs[key] = value


class ThreadLocals(object):
    LOCALS = None

    @staticmethod
    def get_instance():
        if not ThreadLocals.LOCALS:
            ThreadLocals.LOCALS = threadlocal()
        return ThreadLocals()

    def get_item(self, key, default=None):
        return getattr(ThreadLocals.LOCALS, key, default)

    def set_item(self, key, value):
        return setattr(ThreadLocals.LOCALS, key, value)

    def has_item(self, key):
        return hasattr(ThreadLocals.LOCALS, key)

    def del_item(self, key):
        return delattr(ThreadLocals.LOCALS, key)



from functools import wraps
import logging

from six import u

from bugsnag.handlers import BugsnagHandler
from bugsnag import Client
import bugsnag

from tests.utils import IntegrationTest, ScaryException


def use_client_logger(func):
    @wraps(func)
    def wrapped(obj):
        client = Client(use_ssl=False,
                        endpoint=obj.server.address,
                        api_key='tomatoes',
                        asynchronous=False)
        handler = client.log_handler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)
        try:
            func(obj, handler, logger)
        finally:
            logger.removeHandler(handler)

    return wrapped


class HandlersTest(IntegrationTest):

    def setUp(self):
        super(HandlersTest, self).setUp()
        bugsnag.configure(use_ssl=False,
                          endpoint=self.server.address,
                          api_key='tomatoes',
                          notify_release_stages=['dev'],
                          release_stage='dev',
                          asynchronous=False)
        bugsnag.logger.setLevel(logging.INFO)

    def tearDown(self):
        super(HandlersTest, self).tearDown()
        bugsnag.logger.setLevel(logging.CRITICAL)

    def test_message(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.critical('The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('The system is down', exception['message'])

    def test_severity_critical(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.critical('The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogCRITICAL', exception['errorClass'])
        self.assertEqual('error', event['severity'])
        self.assertEqual(logging.CRITICAL,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual(u('CRITICAL'),
                         event['metaData']['extra data']['levelname'])

    def test_severity_error(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.error('The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogERROR', exception['errorClass'])
        self.assertEqual('error', event['severity'])
        self.assertEqual(logging.ERROR,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual(u('ERROR'),
                         event['metaData']['extra data']['levelname'])

    def test_severity_warning(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.warning('The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogWARNING', exception['errorClass'])
        self.assertEqual('warning', event['severity'])
        self.assertEqual(logging.WARNING,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual(u('WARNING'),
                         event['metaData']['extra data']['levelname'])

    def test_severity_info(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.info('The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogINFO', exception['errorClass'])
        self.assertEqual('info', event['severity'])
        self.assertEqual(logging.INFO,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual(u('INFO'),
                         event['metaData']['extra data']['levelname'])

    def test_levelname_message(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        class MessageFilter(logging.Filter):

            def filter(self, record):
                record.levelname = None
                return True

        handler.addFilter(MessageFilter())
        logger.info('The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogMessage', exception['errorClass'])

    def test_custom_level(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.log(341, 'The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogLevel 341', exception['errorClass'])

    def test_custom_levelname(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)
        logging.addLevelName(402, 'OMG')

        logger.log(402, 'The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogOMG', exception['errorClass'])
        self.assertEqual('error', event['severity'])

    def test_exc_info_api_key(self):
        handler = BugsnagHandler(api_key='new news')
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        try:
            raise ScaryException('Oh no')
        except Exception:
            logger.exception('The system is down')

        logger.removeHandler(handler)

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        headers = self.server.received[0]['headers']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('new news', headers['Bugsnag-Api-Key'])
        self.assertEqual(exception['errorClass'], 'tests.utils.ScaryException')

    def test_extra_fields(self):
        handler = BugsnagHandler(api_key='new news',
                                 extra_fields={'fruit': ['grapes', 'pears']})
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.error('A wild tomato appeared', extra={
            'grapes': 8, 'pears': 2, 'tomatoes': 1
        })
        logger.removeHandler(handler)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertEqual(event['metaData']['fruit'], {
            'grapes': 8, 'pears': 2
        })

    def test_client_metadata_fields(self):
        client = Client(use_ssl=False,
                        endpoint=self.server.address,
                        api_key='new news',
                        asynchronous=False)
        handler = client.log_handler(extra_fields={
            'fruit': ['grapes', 'pears']
        })
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.error('A wild tomato appeared', extra={
            'grapes': 8, 'pears': 2, 'tomatoes': 1
        })
        logger.removeHandler(handler)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertEqual(event['metaData']['fruit'], {
            'grapes': 8, 'pears': 2
        })

    @use_client_logger
    def test_client_message(self, handler, logger):
        logger.critical('The system is down')
        self.assertSentReportCount(1)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual('The system is down', exception['message'])

    @use_client_logger
    def test_client_severity_critical(self, handler, logger):
        logger.critical('The system is down')

        self.assertSentReportCount(1)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual('LogCRITICAL', exception['errorClass'])
        self.assertEqual('error', event['severity'])
        self.assertEqual(logging.CRITICAL,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual(u('CRITICAL'),
                         event['metaData']['extra data']['levelname'])

    @use_client_logger
    def test_client_severity_error(self, handler, logger):
        logger.error('The system is down')

        self.assertSentReportCount(1)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual('LogERROR', exception['errorClass'])
        self.assertEqual('error', event['severity'])
        self.assertEqual(logging.ERROR,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual(u('ERROR'),
                         event['metaData']['extra data']['levelname'])

    @use_client_logger
    def test_client_severity_warning(self, handler, logger):
        logger.warning('The system is down')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual('LogWARNING', exception['errorClass'])
        self.assertEqual('warning', event['severity'])
        self.assertEqual(logging.WARNING,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual(u('WARNING'),
                         event['metaData']['extra data']['levelname'])

    @use_client_logger
    def test_client_severity_info(self, handler, logger):
        logger.info('The system is down')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual('LogINFO', exception['errorClass'])
        self.assertEqual('info', event['severity'])
        self.assertEqual(logging.INFO,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual(u('INFO'),
                         event['metaData']['extra data']['levelname'])

    @use_client_logger
    def test_client_add_callback(self, handler, logger):

        def some_callback(record, options):
            for key in record.meals:
                options['meta_data'][key] = record.meals[key]

        handler.add_callback(some_callback)
        logger.info('Everything is fine', extra={'meals': {
            'food': {'fruit': ['pear', 'grape']},
            'drinks': {'free': 'water'}
        }})

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertEqual(event['metaData']['food'], {
            'fruit': ['pear', 'grape']
        })
        self.assertEqual(event['metaData']['drinks'], {
            'free': 'water'
        })

    @use_client_logger
    def test_client_remove_callback(self, handler, logger):

        def some_callback(record, options):
            options['meta_data']['tab'] = {'key': 'value'}

        def some_other_callback(record, options):
            options['meta_data']['tab2'] = {'key': 'value'}

        handler.add_callback(some_callback)
        handler.add_callback(some_other_callback)
        handler.remove_callback(some_callback)
        logger.info('Everything is fine')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertTrue('tab' not in event['metaData'])
        self.assertEqual(event['metaData']['tab2'], {
            'key': 'value'
        })

    @use_client_logger
    def test_client_clear_callbacks(self, handler, logger):

        def some_callback(record, options):
            options['meta_data']['tab'] = {'key': 'value'}

        def some_other_callback(record, options):
            options['meta_data']['tab2'] = {'key': 'value'}

        handler.add_callback(some_callback)
        handler.add_callback(some_other_callback)
        handler.clear_callbacks()
        logger.info('Everything is fine')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertTrue('tab' not in event['metaData'])
        self.assertTrue('tab2' not in event['metaData'])

    @use_client_logger
    def test_client_crashing_callback(self, handler, logger):

        def some_callback(record, options):
            options['meta_data']['tab'] = {'key': 'value'}
            raise ScaryException('Oh dear')

        def some_other_callback(record, options):
            options['meta_data']['tab']['key2'] = 'other value'

        handler.add_callback(some_callback)
        handler.add_callback(some_other_callback)
        logger.info('Everything is fine')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertEqual(event['metaData']['tab'], {
            'key': 'value', 'key2': 'other value'
        })

    @use_client_logger
    def test_client_callback_exception(self, handler, logger):

        def exception_replacing_callback(record, options):
            options['exception'] = ScaryException('replacement')

        handler.add_callback(exception_replacing_callback)
        logger.info('Everything is fine')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual(exception['errorClass'], 'tests.utils.ScaryException')
        self.assertEqual(exception['message'], 'replacement')

    @use_client_logger
    def test_client_callback_exception_metadata(self, handler, logger):

        def exception_replacing_callback(record, options):
            options['exception'] = 'metadata'

        handler.add_callback(exception_replacing_callback)
        logger.info('Everything is fine')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual(exception['errorClass'], 'LogINFO')
        self.assertEqual(exception['message'], 'Everything is fine')
        self.assertEqual(event['metaData']['custom'],
                         {'exception': 'metadata'})



import time

from bugsnag import Client
from bugsnag.notification import Notification
from bugsnag.configuration import Configuration
from bugsnag.sessiontracker import SessionTracker
from bugsnag.utils import ThreadLocals, package_version
from tests.utils import IntegrationTest


class TestConfiguration(IntegrationTest):
    def setUp(self):
        super(TestConfiguration, self).setUp()
        self.config = Configuration()
        self.config.auto_capture_sessions = True

    def test_session_tracker_adds_session_object_to_queue(self):
        tracker = SessionTracker(self.config)
        tracker.auto_sessions = True
        tracker.start_session()
        self.assertEqual(len(tracker.session_counts), 1)
        for key, value in tracker.session_counts.items():
            self.assertEqual(value, 1)

    def test_session_tracker_stores_session_in_threadlocals(self):
        locs = ThreadLocals.get_instance()
        tracker = SessionTracker(self.config)
        tracker.auto_sessions = True
        tracker.start_session()
        session = locs.get_item('bugsnag-session')
        self.assertTrue('id' in session)
        self.assertTrue('startedAt' in session)
        self.assertTrue('events' in session)
        self.assertTrue('handled' in session['events'])
        self.assertTrue('unhandled' in session['events'])
        self.assertEqual(session['events']['handled'], 0)
        self.assertEqual(session['events']['unhandled'], 0)

    def test_session_tracker_sessions_are_unique(self):
        tracker = SessionTracker(self.config)
        tracker.auto_sessions = True
        locs = ThreadLocals.get_instance()
        tracker.start_session()
        session_one = locs.get_item('bugsnag-session').copy()
        tracker.start_session()
        session_two = locs.get_item('bugsnag-session').copy()
        self.assertNotEqual(session_one['id'], session_two['id'])

    def test_session_tracker_send_sessions_sends_sessions(self):
        client = Client(
            auto_capture_sessions=True,
            session_endpoint=self.server.url,
            asynchronous=False
        )
        client.session_tracker.start_session()
        self.assertEqual(len(client.session_tracker.session_counts), 1)
        client.session_tracker.send_sessions()
        self.assertEqual(len(client.session_tracker.session_counts), 0)
        json_body = self.server.received[0]['json_body']
        self.assertTrue('app' in json_body)
        self.assertTrue('notifier' in json_body)
        self.assertTrue('device' in json_body)
        self.assertTrue('sessionCounts' in json_body)
        self.assertEqual(len(json_body['sessionCounts']), 1)

    def test_session_tracker_sets_details_from_config(self):
        client = Client(
            auto_capture_sessions=True,
            session_endpoint=self.server.url,
            asynchronous=False
        )
        client.session_tracker.start_session()
        client.session_tracker.send_sessions()
        json_body = self.server.received[0]['json_body']
        # Notifier properties
        notifier = json_body['notifier']
        self.assertTrue('name' in notifier)
        self.assertEqual(notifier['name'], Notification.NOTIFIER_NAME)
        self.assertTrue('url' in notifier)
        self.assertEqual(notifier['url'], Notification.NOTIFIER_URL)
        self.assertTrue('version' in notifier)
        notifier_version = package_version('bugsnag') or 'unknown'
        self.assertEqual(notifier['version'], notifier_version)
        # App properties
        app = json_body['app']
        self.assertTrue('releaseStage' in app)
        self.assertEqual(app['releaseStage'],
                         client.configuration.get('release_stage'))
        self.assertTrue('version' in app)
        self.assertEqual(app['version'],
                         client.configuration.get('app_version'))
        # Device properties
        device = json_body['device']
        self.assertTrue('hostname' in device)
        self.assertEqual(device['hostname'],
                         client.configuration.get('hostname'))

    def test_session_middleware_attaches_session_to_notification(self):
        client = Client(
            auto_capture_sessions=True,
            session_endpoint=self.server.url + '/ignore',
            endpoint=self.server.url,
            asynchronous=False
        )
        client.session_tracker.start_session()
        client.notify(Exception("Test"))
        while len(self.server.received) == 0:
            time.sleep(0.5)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertTrue('session' in event)
        session = event['session']
        self.assertTrue('id' in session)
        self.assertTrue('startedAt' in session)
        self.assertTrue('events' in session)
        sesevents = session['events']
        self.assertTrue('unhandled' in sesevents)
        self.assertEqual(sesevents['unhandled'], 0)
        self.assertTrue('handled' in sesevents)
        self.assertEqual(sesevents['handled'], 1)



import sys
import logging

from bugsnag.configuration import Configuration, RequestConfiguration
from bugsnag.notification import Notification
from bugsnag.client import Client
from bugsnag.legacy import (configuration, configure, configure_request,
                            add_metadata_tab, clear_request_config, notify,
                            auto_notify, before_notify, start_session,
                            send_sessions)

__all__ = ('Client', 'Notification', 'Configuration', 'RequestConfiguration',
           'configuration', 'configure', 'configure_request',
           'add_metadata_tab', 'clear_request_config', 'notify',
           'auto_notify', 'before_notify', 'start_session',
           'send_sessions')

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)
handler = logging.StreamHandler(sys.stderr)
formatter = logging.Formatter(
    '%(asctime)s - [%(name)s] %(levelname)s - %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)



from threading import Thread
import sys
import json

from time import strftime, gmtime

from six.moves.urllib.request import (
    Request,
    ProxyHandler,
    build_opener
)

import bugsnag
from bugsnag.notification import Notification

try:
    if sys.version_info < (2, 7):
        raise ImportError('requests-based delivery will fail on 2.6 '
                          'and earlier')

    if (3, 1) < sys.version_info < (3, 3):
        raise ImportError('requests-based delivery will fail on 3.2')

    import requests
except ImportError:
    requests = None


def create_default_delivery():
    if requests is not None:
        return RequestsDelivery()

    return UrllibDelivery()


def default_headers(api_key):
    return {
        'Bugsnag-Api-Key': api_key,
        'Bugsnag-Payload-Version': Notification.PAYLOAD_VERSION,
        'Bugsnag-Sent-At': strftime('%Y-%m-%dT%H:%M:%S', gmtime()),
        'Content-Type': 'application/json',
    }


class Delivery(object):
    """
    Mechanism for sending a request to Bugsnag
    """

    def deliver(self, config, payload):
        """
        Sends error reports to Bugsnag
        """
        pass

    def deliver_sessions(self, config, payload):
        """
        Sends sessions to Bugsnag
        """
        pass


class UrllibDelivery(Delivery):

    def deliver_sessions(self, config, payload):
        options = {
            'endpoint': config.session_endpoint,
            'success': 202,
        }
        self.deliver(config, payload, options)

    def deliver(self, config, payload, options={}):

        def request():
            uri = options.pop('endpoint', config.endpoint)
            if '://' not in uri:
                uri = config.get_endpoint()
            api_key = json.loads(payload).pop('apiKey', config.get('api_key'))
            req = Request(uri,
                          payload.encode('utf-8', 'replace'),
                          default_headers(api_key))

            if config.proxy_host:
                proxies = ProxyHandler({
                    'https': config.proxy_host,
                    'http': config.proxy_host
                })

                opener = build_opener(proxies)
            else:
                opener = build_opener()

            resp = opener.open(req)
            status = resp.getcode()

            if 'success' in options:
                success = options['success']
            else:
                success = 200
            if status != success:
                bugsnag.logger.warning(
                    'Delivery to %s failed, status %d' % (uri, status))

        if config.asynchronous:
            Thread(target=request).start()
        else:
            request()


class RequestsDelivery(Delivery):

    def deliver_sessions(self, config, payload):
        options = {
            'endpoint': config.session_endpoint,
            'success': 202,
        }
        self.deliver(config, payload, options)

    def deliver(self, config, payload, options={}):

        def request():
            uri = options.pop('endpoint', config.endpoint)
            if '://' not in uri:
                uri = config.get_endpoint()

            api_key = json.loads(payload).pop('apiKey', config.get('api_key'))
            req_options = {'data': payload,
                           'headers': default_headers(api_key)}

            if config.proxy_host:
                req_options['proxies'] = {
                    'https': config.proxy_host,
                    'http': config.proxy_host
                }

            response = requests.post(uri, **req_options)
            status = response.status_code
            if 'success' in options:
                success = options['success']
            else:
                success = requests.codes.ok

            if status != success:
                bugsnag.logger.warning(
                    'Delivery to %s failed, status %d' % (uri, status))

        if config.asynchronous:
            Thread(target=request).start()
        else:
            request()



from flask import Flask
from bugsnag.flask import handle_exceptions
import bugsnag.notification
from tests.utils import IntegrationTest


class SentinelError(RuntimeError):
    pass


class TestFlask(IntegrationTest):

    def setUp(self):
        super(TestFlask, self).setUp()
        bugsnag.configure(use_ssl=False,
                          endpoint=self.server.address,
                          api_key='3874876376238728937',
                          notify_release_stages=['dev'],
                          release_stage='dev',
                          asynchronous=False)

    def test_bugsnag_middleware_working(self):
        app = Flask("bugsnag")

        @app.route("/hello")
        def hello():
            return "OK"

        handle_exceptions(app)

        resp = app.test_client().get('/hello')
        self.assertEqual(resp.data, b'OK')

        self.assertEqual(0, len(self.server.received))

    def test_bugsnag_crash(self):
        app = Flask("bugsnag")

        @app.route("/hello")
        def hello():
            raise SentinelError("oops")

        handle_exceptions(app)
        app.test_client().get('/hello')

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        self.assertEqual(payload['events'][0]['exceptions'][0]['errorClass'],
                         'test_flask.SentinelError')
        self.assertEqual(payload['events'][0]['metaData']['request']['url'],
                         'http://localhost/hello')

    def test_bugsnag_notify(self):
        app = Flask("bugsnag")

        @app.route("/hello")
        def hello():
            bugsnag.notify(SentinelError("oops"))
            return "OK"

        handle_exceptions(app)
        app.test_client().get('/hello')

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        self.assertEqual(payload['events'][0]['metaData']['request']['url'],
                         'http://localhost/hello')

    def test_bugsnag_custom_data(self):
        meta_data = [{"hello": {"world": "once"}},
                     {"again": {"hello": "world"}}]

        app = Flask("bugsnag")

        @app.route("/hello")
        def hello():
            bugsnag.configure_request(meta_data=meta_data.pop())
            raise SentinelError("oops")

        handle_exceptions(app)
        with app.test_client() as client:
            client.get('/hello')
            client.get('/hello')

        payload = self.server.received[0]['json_body']
        event = payload['events'][0]
        self.assertEqual(event['metaData'].get('hello'), None)
        self.assertEqual(event['metaData']['again']['hello'], 'world')

        payload = self.server.received[1]['json_body']
        event = payload['events'][0]
        self.assertEqual(event['metaData']['hello']['world'], 'once')
        self.assertEqual(event['metaData'].get('again'), None)
        self.assertEqual(2, len(self.server.received))

    def test_bugsnag_includes_posted_json_data(self):
        app = Flask("bugsnag")

        @app.route("/ajax", methods=["POST"])
        def hello():
            raise SentinelError("oops")

        handle_exceptions(app)
        app.test_client().post(
            '/ajax', data='{"key": "value"}', content_type='application/json')

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        event = payload['events'][0]
        self.assertEqual(event['exceptions'][0]['errorClass'],
                         'test_flask.SentinelError')
        self.assertEqual(event['metaData']['request']['url'],
                         'http://localhost/ajax')
        self.assertEqual(event['metaData']['request']['data'],
                         dict(key='value'))

    def test_bugsnag_add_metadata_tab(self):
        app = Flask("bugsnag")

        @app.route("/form", methods=["PUT"])
        def hello():
            bugsnag.add_metadata_tab("account", {"id": 1, "premium": True})
            bugsnag.add_metadata_tab("account", {"premium": False})
            raise SentinelError("oops")

        handle_exceptions(app)
        app.test_client().put(
            '/form', data='_data', content_type='application/octet-stream')

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        event = payload['events'][0]
        self.assertEqual(event['metaData']['account']['premium'], False)
        self.assertEqual(event['metaData']['account']['id'], 1)

    def test_bugsnag_includes_unknown_content_type_posted_data(self):
        app = Flask("bugsnag")

        @app.route("/form", methods=["PUT"])
        def hello():
            raise SentinelError("oops")

        handle_exceptions(app)
        app.test_client().put(
            '/form', data='_data', content_type='application/octet-stream')

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        event = payload['events'][0]
        self.assertEqual(event['exceptions'][0]['errorClass'],
                         'test_flask.SentinelError')
        self.assertEqual(event['metaData']['request']['url'],
                         'http://localhost/form')
        body = event['metaData']['request']['data']['body']
        self.assertTrue('_data' in body)

    def test_bugsnag_notify_with_custom_context(self):
        app = Flask("bugsnag")

        @app.route("/hello")
        def hello():
            bugsnag.notify(SentinelError("oops"),
                           context="custom_context_notification_testing")
            return "OK"

        handle_exceptions(app)
        app.test_client().get('/hello')

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        self.assertEqual(payload['events'][0]['context'],
                         'custom_context_notification_testing')

    def test_flask_intergration_includes_middleware_severity(self):
        app = Flask("bugsnag")

        @app.route("/test")
        def test():
            raise SentinelError("oops")

        handle_exceptions(app)
        app.test_client().get("/test")

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        event = payload['events'][0]
        self.assertTrue(event['unhandled'])
        self.assertEqual(event['severityReason'], {
            "type": "unhandledExceptionMiddleware",
            "attributes": {
                "framework": "Flask"
            }
        })



from webtest import TestApp

from bugsnag.wsgi.middleware import BugsnagMiddleware
from six import Iterator
import bugsnag.notification
import bugsnag
from tests.utils import IntegrationTest


class SentinelError(RuntimeError):
    pass


class TestWSGI(IntegrationTest):

    def setUp(self):
        super(TestWSGI, self).setUp()
        bugsnag.configure(use_ssl=False,
                          endpoint=self.server.address,
                          api_key='3874876376238728937',
                          notify_release_stages=['dev'],
                          release_stage='dev',
                          asynchronous=False)

    def test_bugsnag_middleware_working(self):
        def BasicWorkingApp(environ, start_response):
            start_response("200 OK",
                           [('Content-Type', 'text/plain; charset=utf-8')])
            return iter([b'OK'])

        app = TestApp(BugsnagMiddleware(BasicWorkingApp))

        resp = app.get('/', status=200)

        self.assertEqual(resp.body, b'OK')
        self.assertEqual(0, len(self.server.received))

    def test_bugsnag_middleware_crash_on_start(self):

        class CrashOnStartApp(object):
            def __init__(self, environ, start_response):
                raise SentinelError("oops")

        app = TestApp(BugsnagMiddleware(CrashOnStartApp))

        self.assertRaises(SentinelError, lambda: app.get('/beans'))

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        event = payload['events'][0]
        self.assertEqual(event['context'], 'GET /beans')
        self.assertEqual(event['metaData']['environment']['PATH_INFO'],
                         '/beans')

    def test_bugsnag_middleware_crash_on_iter(self):
        class CrashOnIterApp(Iterator):
            def __init__(self, environ, start_response):
                pass

            def __iter__(self):
                return self

            def __next__(self):
                raise SentinelError("oops")
        app = TestApp(BugsnagMiddleware(CrashOnIterApp))

        self.assertRaises(SentinelError, lambda: app.get('/beans'))

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        environ = payload['events'][0]['metaData']['environment']
        self.assertEqual(environ['PATH_INFO'], '/beans')

    def test_bugsnag_middleware_crash_on_close(self):
        class CrashOnCloseApp(Iterator):
            def __init__(self, environ, start_response):
                pass

            def __iter__(self):
                return self

            def __next__(self):
                raise StopIteration()

            def close(self):
                raise SentinelError("oops")

        app = TestApp(BugsnagMiddleware(CrashOnCloseApp))

        self.assertRaises(SentinelError, lambda: app.get('/beans'))

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        environ = payload['events'][0]['metaData']['environment']
        self.assertEqual(environ['PATH_INFO'], '/beans')

    def test_bugsnag_middleware_respects_user_id(self):

        class CrashAfterSettingUserId(object):
            def __init__(self, environ, start_response):
                bugsnag.configure_request(user={
                    "id": "5",
                    "email": "me@cirw.in",
                    "name": "conrad",
                })
                raise SentinelError("oops")

        app = TestApp(BugsnagMiddleware(CrashAfterSettingUserId))

        self.assertRaises(SentinelError, lambda: app.get('/beans'))

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        self.assertEqual(payload['events'][0]['user']['id'], '5')

    def test_bugsnag_middleware_respects_meta_data(self):

        class CrashAfterSettingMetaData(object):
            def __init__(self, environ, start_response):
                bugsnag.configure_request(meta_data={"account":
                                                     {"paying": True}})

            def __iter__(self):
                raise SentinelError("oops")

        app = TestApp(BugsnagMiddleware(CrashAfterSettingMetaData))

        self.assertRaises(SentinelError, lambda: app.get('/beans'))

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        event = payload['events'][0]
        self.assertEqual(event['metaData']['account'], {"paying": True})

    def test_bugsnag_middleware_closes_iterables(self):

        class CrashOnCloseIterable(object):

            def __init__(self, environ, start_response):
                start_response("200 OK",
                               [('Content-Type', 'text/plain; charset=utf-8')])

            def __iter__(self):
                yield 'OK'

            def close(self):
                raise SentinelError("oops")

        app = TestApp(BugsnagMiddleware(CrashOnCloseIterable))

        self.assertRaises(SentinelError, lambda: app.get('/beans'))

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        environ = payload['events'][0]['metaData']['environment']
        self.assertEqual(environ['PATH_INFO'], '/beans')

    def test_bugsnag_middleware_attaches_unhandled_data(self):

        class CrashOnStartApp(object):

            def __init__(self, environ, start_response):
                raise SentinelError("oops")

        app = TestApp(BugsnagMiddleware(CrashOnStartApp))

        self.assertRaises(SentinelError, lambda: app.get('/beans'))

        self.assertEqual(1, len(self.server.received))
        payload = self.server.received[0]['json_body']
        event = payload['events'][0]

        self.assertTrue(event['unhandled'])
        self.assertEqual(event['severityReason'], {
            'type': 'unhandledExceptionMiddleware',
            'attributes': {
                'framework': 'WSGI'
            }
        })



"""
This example shows how to send exceptions in celery tasks to Bugsnag.

First:

    Change the API key below to your API key,

Then, in one terminal, run:

    celery -A example.tasks

And, in another, run:

    python -m example.tasks

You will see a zero-division error appear in Bugsnag.
"""

from celery import Celery
import bugsnag
from bugsnag.celery import connect_failure_handler

celery = Celery('tasks', broker='redis://localhost', backend='redis')

bugsnag.configure(api_key="066f5ad3590596f9aa8d601ea89af845")
connect_failure_handler()

@celery.task(name='tasks.divide')
def divide(x, y):
    return x / y

if __name__ == "__main__":
    divide.delay(1, 0)
